@@ -44,19 +44,68 @@ def make_am(msgs: List[email.message.Message], cmdargs: argparse.Namespace, msgi
     if outdir == '-':
         cmdargs.nocover = True
     wantver = cmdargs.wantver
-    covertrailers = cmdargs.covertrailers
     count = len(msgs)
     logger.info('Analyzing %s messages in the thread', count)
     lmbx = b4.LoreMailbox()
     # Go through the mbox once to populate base series
+    load_codereview = True
     for msg in msgs:
-        lmbx.add_message(msg)
+        # Is it a collection of patches attached to the same message?
+        # We only trigger this mode with --single-msg
+        if msg.is_multipart() and ('singlemsg' in cmdargs and cmdargs.singlemsg):
+            xpatches = list()
+            xmsgid = b4.LoreMessage.get_clean_msgid(msg)
+            for part in msg.walk():
+                cte = part.get_content_type()
+                if cte.find('/x-patch') < 0:
+                    continue
+                payload = part.get_payload(decode=True)
+                if payload is None:
+                    continue
+                pcharset = part.get_content_charset()
+                if not pcharset:
+                    pcharset = 'utf-8'
+                try:
+                    payload = payload.decode(pcharset, errors='replace')
+                except LookupError:
+                    # what kind of encoding is that?
+                    # Whatever, we'll use utf-8 and hope for the best
+                    payload = payload.decode('utf-8', errors='replace')
+                    part.set_param('charset', 'utf-8')
+                if payload and b4.DIFF_RE.search(payload):
+                    xmsg = email.message_from_string(payload, policy=b4.emlpolicy)
+                    # Needs to have Subject, From, Date for us to consider it
+                    if xmsg.get('Subject') and xmsg.get('From') and xmsg.get('Date'):
+                        logger.debug('Found attached patch: %s', xmsg.get('Subject'))
+                        xmsg['Message-ID'] = f'<att{len(xpatches)}-{xmsgid}>'
+                        xpatches.append(xmsg)
+            if len(xpatches):
+                logger.info('Warning: Found %s patches attached to the requested message', len(xpatches))
+                logger.info('         This mode ignores any follow-up trailers, use with caution')
+                # Throw out lmbx and only use these
+                lmbx = b4.LoreMailbox()
+                load_codereview = False
+                for xmsg in xpatches:
+                    lmbx.add_message(xmsg)
+                # Make a cover letter out of the original message
+                cmsg = email.message.EmailMessage()
+                cbody, ccharset = b4.LoreMessage.get_payload(msg, use_patch=False)
+                cmsg['From'] = msg.get('From')
+                cmsg['Date'] = msg.get('Date')
+                cmsg['Message-ID'] = msg.get('Message-ID')
+                cmsg['Subject'] = '[PATCH 0/0] ' + msg.get('Subject')
+                cmsg.set_payload(cbody, ccharset)
+                lmbx.add_message(cmsg)
+                break
+        else:
+            lmbx.add_message(msg)
 
     reroll = True
     if cmdargs.nopartialreroll:
         reroll = False
 
-    lser = lmbx.get_series(revision=wantver, sloppytrailers=cmdargs.sloppytrailers, reroll=reroll)
+    lser = lmbx.get_series(revision=wantver, sloppytrailers=cmdargs.sloppytrailers, reroll=reroll,
+                           codereview_trailers=load_codereview)
     if lser is None and cmdargs.cherrypick != '_':
         if wantver is None:
             logger.critical('No patches found.')
@@ -99,18 +148,12 @@ def make_am(msgs: List[email.message.Message], cmdargs: argparse.Namespace, msgi
                 logger.critical('Could not match "%s" to any subjects in the series', cmdargs.cherrypick)
                 sys.exit(1)
         else:
-            cherrypick = list(b4.parse_int_range(cmdargs.cherrypick, upper=len(lser.patches)-1))
+            cherrypick = list(b4.parse_int_range(cmdargs.cherrypick, upper=len(lser.patches) - 1))
     else:
         cherrypick = None
 
-    try:
-        am_msgs = lser.get_am_ready(noaddtrailers=cmdargs.noaddtrailers,
-                                    covertrailers=covertrailers, addmysob=cmdargs.addmysob,
-                                    addlink=cmdargs.addlink, linkmask=config['linkmask'], cherrypick=cherrypick,
-                                    copyccs=cmdargs.copyccs, allowbadchars=cmdargs.allowbadchars)
-    except KeyError:
-        sys.exit(1)
-
+    am_msgs = lser.get_am_ready(noaddtrailers=cmdargs.noaddtrailers, addmysob=cmdargs.addmysob, addlink=cmdargs.addlink,
+                                cherrypick=cherrypick, copyccs=cmdargs.copyccs, allowbadchars=cmdargs.allowbadchars)
     logger.info('---')
 
     if cherrypick is None:
@@ -118,16 +161,6 @@ def make_am(msgs: List[email.message.Message], cmdargs: argparse.Namespace, msgi
     else:
         logger.info('Total patches: %s (cherrypicked: %s)', len(am_msgs), cmdargs.cherrypick)
 
-    if lser.has_cover and lser.patches[0].followup_trailers and not covertrailers:
-        # Warn that some trailers were sent to the cover letter
-        logger.critical('---')
-        logger.critical('NOTE: Some trailers were sent to the cover letter:')
-        tseen = set()
-        for ltr in lser.patches[0].followup_trailers:
-            if ltr not in tseen:
-                logger.critical('      %s', ltr.as_string(omit_extinfo=True))
-                tseen.add(ltr)
-        logger.critical('NOTE: Rerun with -t to apply them to all patches')
     if len(lser.trailer_mismatches):
         logger.critical('---')
         logger.critical('NOTE: some trailers ignored due to from/email mismatches:')
@@ -157,7 +190,7 @@ def make_am(msgs: List[email.message.Message], cmdargs: argparse.Namespace, msgi
         else:
             rstart, rend = lser.make_fake_am_range(gitdir=None)
             if rstart and rend:
-                logger.info('Prepared a fake commit range for 3-way merge (%.12s..%.12s)', rstart, rend)
+                logger.info('Preared a fake commit range for 3-way merge (%.12s..%.12s)', rstart, rend)
 
     logger.critical('---')
     if lser.partial_reroll:
@@ -218,6 +251,7 @@ def make_am(msgs: List[email.message.Message], cmdargs: argparse.Namespace, msgi
         logger.critical(' Link: %s', linkurl)
 
     base_commit = None
+
     matches = re.search(r'base-commit: .*?([\da-f]+)', first_body, re.MULTILINE)
     if matches:
         base_commit = matches.groups()[0]
@@ -232,7 +266,7 @@ def make_am(msgs: List[email.message.Message], cmdargs: argparse.Namespace, msgi
         if not b4.git_commit_exists(topdir, base_commit):
             logger.info(' Base: base-commit %s not known, ignoring', base_commit)
             base_commit = None
-        else:
+        elif not cmdargs.mergebase:
             logger.info(' Base: using specified base-commit %s', base_commit)
 
     if not base_commit and topdir and cmdargs.guessbase:
@@ -250,6 +284,12 @@ def make_am(msgs: List[email.message.Message], cmdargs: argparse.Namespace, msgi
         except IndexError:
             logger.critical(' Base: failed to guess base')
 
+    if cmdargs.mergebase:
+        if base_commit:
+            logger.warn(' Base: overriding submitter provided base-commit %s', base_commit)
+        base_commit = cmdargs.mergebase
+        logger.info(' Base: using CLI provided base-commit %s', base_commit)
+
     if cmdargs.subcmd == 'shazam':
         if not topdir:
             logger.critical('Could not figure out where your git dir is, cannot shazam.')
@@ -272,99 +312,59 @@ def make_am(msgs: List[email.message.Message], cmdargs: argparse.Namespace, msgi
             # Try our best with HEAD, I guess
             base_commit = 'HEAD'
 
-        with b4.git_temp_worktree(topdir, base_commit) as gwt:
-            logger.info('Magic: Preparing a sparse worktree')
-            ecode, out = b4.git_run_command(gwt, ['sparse-checkout', 'init'], logstderr=True)
-            if ecode > 0:
-                logger.critical('Error running sparse-checkout init')
-                logger.critical(out)
-                sys.exit(ecode)
-            ecode, out = b4.git_run_command(gwt, ['checkout'], logstderr=True)
-            if ecode > 0:
-                logger.critical('Error running checkout into sparse workdir')
-                logger.critical(out)
-                sys.exit(ecode)
-            ecode, out = b4.git_run_command(gwt, ['am'], stdin=ambytes, logstderr=True)
-            if ecode > 0:
-                logger.critical('Unable to cleanly apply series, see failure log below')
-                logger.critical('---')
-                logger.critical(out.strip())
-                logger.critical('---')
-                logger.critical('Not fetching into FETCH_HEAD')
-                sys.exit(ecode)
-            logger.info('---')
-            logger.info(out.strip())
-            logger.info('---')
-            logger.info('Fetching into FETCH_HEAD')
-            gitargs = ['fetch', gwt]
-            ecode, out = b4.git_run_command(topdir, gitargs, logstderr=True)
-            if ecode > 0:
-                logger.critical('Unable to fetch from the worktree')
-                logger.critical(out.strip())
-                sys.exit(ecode)
-            gitargs = ['rev-parse', '--git-path', 'FETCH_HEAD']
-            ecode, fhf = b4.git_run_command(topdir, gitargs, logstderr=True)
-            if ecode > 0:
-                logger.critical('Unable to find FETCH_HEAD')
-                logger.critical(out.strip())
-                sys.exit(ecode)
-            with open(fhf.rstrip(), 'r') as fhh:
-                contents = fhh.read()
-            linkurl = config['linkmask'] % top_msgid
-            if len(am_msgs) > 1:
-                mmsg = 'patches from %s' % linkurl
-            else:
-                mmsg = 'patch from %s' % linkurl
-            new_contents = contents.replace(gwt, mmsg)
-            if new_contents != contents:
-                with open(fhf, 'w') as fhh:
-                    fhh.write(new_contents)
+        linkurl = config['linkmask'] % top_msgid
+        try:
+            b4.git_fetch_am_into_repo(topdir, am_msgs=am_msgs, at_base=base_commit, origin=linkurl)
+        except RuntimeError:
+            sys.exit(1)
 
-            gitargs = ['rev-parse', '--git-dir']
-            ecode, fhf = b4.git_run_command(topdir, gitargs, logstderr=True)
-            if ecode > 0:
-                logger.critical('Unable to find git directory')
-                logger.critical(out.strip())
-                sys.exit(ecode)
-            mmf = os.path.join(fhf.rstrip(), 'b4-cover')
-            merge_template = DEFAULT_MERGE_TEMPLATE
-            if config.get('shazam-merge-template'):
-                # Try to load this template instead
-                try:
-                    merge_template = b4.read_template(config['shazam-merge-template'])
-                except FileNotFoundError:
-                    logger.critical('ERROR: shazam-merge-template says to use %s, but it does not exist',
-                                    config['shazam-merge-template'])
-                    sys.exit(2)
+        gitargs = ['rev-parse', '--git-dir']
+        ecode, out = b4.git_run_command(topdir, gitargs, logstderr=True)
+        if ecode > 0:
+            logger.critical('Unable to find git directory')
+            logger.critical(out.strip())
+            sys.exit(ecode)
+        mmf = os.path.join(out.rstrip(), 'b4-cover')
+        merge_template = DEFAULT_MERGE_TEMPLATE
+        if config.get('shazam-merge-template'):
+            # Try to load this template instead
+            try:
+                merge_template = b4.read_template(config['shazam-merge-template'])
+            except FileNotFoundError:
+                logger.critical('ERROR: shazam-merge-template says to use %s, but it does not exist',
+                                config['shazam-merge-template'])
+                sys.exit(2)
 
-            # Write out a sample merge message using the cover letter
-            if os.path.exists(mmf):
-                # Make sure any old cover letters don't confuse anyone
-                os.unlink(mmf)
+        # Write out a sample merge message using the cover letter
+        if os.path.exists(mmf):
+            # Make sure any old cover letters don't confuse anyone
+            os.unlink(mmf)
 
-            if lser.has_cover:
-                cmsg = lser.patches[0]
-                parts = b4.LoreMessage.get_body_parts(cmsg.body)
-                covermessage = parts[1]
-            else:
-                cmsg = lser.patches[1]
-                covermessage = ('NOTE: No cover letter provided by the author.\n'
-                                '      Add merge commit message here.')
-            tptvals = {
-                'seriestitle': cmsg.subject,
-                'authorname': cmsg.fromname,
-                'authoremail': cmsg.fromemail,
-                'covermessage': covermessage,
-                'midurl': linkurl,
-            }
-            if len(am_msgs) > 1:
-                tptvals['patch_or_series'] = 'patch series'
-            else:
-                tptvals['patch_or_series'] = 'patch'
+        if lser.has_cover:
+            cmsg = lser.patches[0]
+            parts = b4.LoreMessage.get_body_parts(cmsg.body)
+            covermessage = parts[1]
+        else:
+            cmsg = lser.patches[1]
+            covermessage = ('NOTE: No cover letter provided by the author.\n'
+                            '      Add merge commit message here.')
 
-            body = Template(merge_template).safe_substitute(tptvals)
-            with open(mmf, 'w') as mmh:
-                mmh.write(body)
+        tptvals = {
+            'seriestitle': cmsg.subject,
+            'authorname': cmsg.fromname,
+            'authoremail': cmsg.fromemail,
+            'covermessage': covermessage,
+            'mid': top_msgid,
+            'midurl': linkurl,
+        }
+        if len(am_msgs) > 1:
+            tptvals['patch_or_series'] = 'patch series'
+        else:
+            tptvals['patch_or_series'] = 'patch'
+
+        body = Template(merge_template).safe_substitute(tptvals)
+        with open(mmf, 'w') as mmh:
+            mmh.write(body)
 
         mergeflags = config.get('shazam-merge-flags', '--signoff')
         sp = shlex.shlex(mergeflags, posix=True)
@@ -410,7 +410,7 @@ def make_am(msgs: List[email.message.Message], cmdargs: argparse.Namespace, msgi
     if base_commit is not None:
         logger.critical('       git checkout -b %s %s', gitbranch, base_commit)
     if cmdargs.outdir != '-':
-        logger.critical('       git am %s', am_filename)
+        logger.critical('       git am %s%s', '-3 ' if cmdargs.threeway else '', am_filename)
     thanks_record_am(lser, cherrypick=cherrypick)
 
 
@@ -676,8 +676,8 @@ def refetch(dest: str) -> None:
 
 
 def main(cmdargs: argparse.Namespace) -> None:
+    # We force some settings
     if cmdargs.subcmd == 'shazam':
-        # We force some settings
         cmdargs.checknewer = True
         cmdargs.threeway = False
         cmdargs.nopartialreroll = False
@@ -689,6 +689,8 @@ def main(cmdargs: argparse.Namespace) -> None:
             cmdargs.guessbase = True
         else:
             cmdargs.guessbase = False
+    else:
+        cmdargs.mergebase = False
 
     if cmdargs.checknewer:
         # Force nocache mode
